
# initialize libraries.
import os
import numpy
import logging
from scipy.special import psi
from scipy.stats.distributions import norm
from operator import attrgetter, itemgetter
import multiprocessing
from multiprocessing import sharedctypes
import itertools
import time
import gc 

from .classDef import Peak

debug = logging.debug
info = logging.info

def get_candidate_window2( x, y, repx, repy, threshold):
    # using PHI = 1e6 to prescreen the genome 
    PHI = 1e6
    GAMMA_HAT = 1.0
    tau = numpy.sqrt( y*( repx*x*(PHI+y) + repy*y*(PHI+x))/repx/repy/PHI/x**3)
    gamma = y/x
    z = (numpy.log(gamma)-numpy.log(GAMMA_HAT))*gamma/tau
    pvalue = norm.cdf(-z)
    pre_idx_list = numpy.where(pvalue[10:-10]<threshold)[0]+10
    return numpy.array(pre_idx_list)


def weighted_log_likelihood(v_hat, m, n, reads, diff_test):
    '''This function returns the derivative of the log likelihood
       of current and adjacent windows using a triangular weight.'''
    equation = 0
    n_window = len(reads)
    baseline = numpy.mean(reads[int(n_window/2),0:m])
    for idx in range(n_window):
        x = reads[idx, 0:m]  # x/m refer the test sample
        y = reads[idx, m:(m+n)]  # y/n refer to the control sample
        weight_x = numpy.mean(x)/baseline
        weight_y = numpy.mean(y)/baseline
        if n == 1: 
            weight_y = 0
        log_likelihood = (-(m*weight_x+n*weight_y)*psi(v_hat) +
                numpy.sum(psi(v_hat+x))*weight_x + 
                numpy.sum(psi(v_hat+y))*weight_y + 
                m*numpy.log(v_hat/(v_hat+numpy.mean(x)))*weight_x + 
                n*numpy.log(v_hat/(v_hat+numpy.mean(y)))*weight_y)
        equation = (equation + 
                log_likelihood*(1-(abs(float(n_window)/2-idx-0.5)/(float(n_window)/2+1))))
    return equation


def estimate_area_dispersion_factor(read, m, n, idx, peaktype, diff_test):
    '''This function implements a simple method to find the root
       of the derivative of the log likelihood.'''
    BASE = 10
    N_ITERATION = 15
    if peaktype == "sharp":
        N_WINDOWS = 1 
    elif peaktype  == "broad":
        N_WINDOWS = 5
    left = -1.0
    right = 6.0
    if weighted_log_likelihood(BASE**right, m, n, 
            read[(idx-N_WINDOWS):(idx+N_WINDOWS+1)], diff_test) > 0:
        return BASE**right
    elif weighted_log_likelihood(BASE**left, m, n,
            read[(idx-N_WINDOWS):(idx+N_WINDOWS+1)], diff_test) < 0:
        return BASE**left
    for times in range(N_ITERATION):
        temp = (right+left)/2
        if weighted_log_likelihood(10**temp, m, n,
                read[(idx-N_WINDOWS):(idx+N_WINDOWS+1)], diff_test) > 0:
            left = temp
        else:
            right = temp
    return BASE**right


def cal_FDR(peak_list, num_tests):
    '''Calculate the Benjamini-Hochberg FDR'''
    if len(peak_list)==0:
        return peak_list
    peak_list = sorted(peak_list, key=attrgetter('pvalue'))
    #calculate BH q-values
    q_list = [item.pvalue*num_tests/(idx+1) 
            for idx, item in enumerate(peak_list)]
    q_min = min(q_list[-1],1) # q-value should be less than 1. 
    for i in range(len(q_list)-1, -1, -1):
        if q_list[i] >= q_min:
            q_list[i] = q_min
        else:
            q_min = q_list[i]
    for i in range(len(q_list)):
        peak_list[i] = Peak(peak_list[i].chr, peak_list[i].index,
                            peak_list[i].g1_count, peak_list[i].g2_count,
                            peak_list[i].pvalue, q_list[i])
    peak_list = sorted(peak_list, key=attrgetter('chr', 'index'))
    return peak_list


    
def per_chr_nbtest(read_array, chr, swap,threshold, peaktype,difftest, start1,end1,start2,end2,test_rep,control_rep):
    t1 = time.time()
    sig_peaks_list = []
    y_bar_array = numpy.mean(read_array[:, start1:end1], 1)
    x_bar_array = numpy.mean(read_array[:, start2:end2], 1)
    if swap: #swap the chip and control reads.
        x_bar_array, y_bar_array = y_bar_array, x_bar_array
    cand_index = get_candidate_window2( x_bar_array,
                    y_bar_array, control_rep, test_rep, threshold)
    debug("There are %d candidate windows for %s (PID:%d)", len(cand_index), chr, os.getpid())
    if not swap:
        disp_list = numpy.array([estimate_area_dispersion_factor(read_array,
                test_rep, control_rep, idx, peaktype, difftest)
                for idx in cand_index])
    else:
        disp_list = numpy.array([estimate_area_dispersion_factor(read_array,
                control_rep, test_rep, idx, peaktype, difftest)
                for idx in cand_index])
    #debug("finished estimating dispersion for %s", chr)
   # return []
    cand_x_bar_array = x_bar_array[cand_index]
    cand_y_bar_array = y_bar_array[cand_index]
    gamma_array = cand_y_bar_array / cand_x_bar_array
    tau_hat_array = numpy.sqrt(cand_y_bar_array*
            ((control_rep*cand_x_bar_array*(disp_list+cand_y_bar_array)) +
            (test_rep*cand_y_bar_array*(disp_list+cand_x_bar_array)))/
            (test_rep*control_rep*disp_list*(cand_x_bar_array**3)))

    gamma_hat = 1.0 #Null hypothesis
    z_score_array = ((numpy.log(gamma_array)-numpy.log(gamma_hat))*
            gamma_array/tau_hat_array)
    pval_array = norm.cdf(-z_score_array)
    test_index = numpy.where(pval_array<threshold)
    test_index = test_index[0]
    sig_index = cand_index[test_index]
    sig_pval = pval_array[test_index]
    sig_group1_count = cand_y_bar_array[test_index]
    sig_group2_count = cand_x_bar_array[test_index]
    #sig_disp = disp_list[test_index]
    for i, a in enumerate(test_index):
        sig_peaks_list.append(Peak(chr, sig_index[i], sig_group1_count[i], sig_group2_count[i], sig_pval[i], 0))
    t2 = time.time()
    debug ("Analysis finished for %s, used %f sec CPU time", chr, t2-t1)
    return sig_peaks_list
	
def per_chr_nbtest_helper(args):
	### help the workers to take multiple arguments 
    try: 
        return per_chr_nbtest(*args)
    except KeyboardInterrupt:
        pass 

def mock_helper(args):
    return mock(*args)

def mock(d,p):
    return []        
 
def negative_binomial(read_dict,peakfilename, swap, parameter):
    '''the main function that test for significant windows.'''
    # Initialize the parameters
    peaktype = parameter.peaktype
    threshold = parameter.threshold
    windowsize = parameter.window_size
    # Indicate the data 
    if parameter.difftest is True: 
        test_list = parameter.chip1
        control_list = parameter.chip2
    else: 
        test_list = parameter.chip1
        control_list = parameter.input1
    num_tests = parameter.get_genome_size()/windowsize
    

    #compute number of replicates
    test_rep = len(test_list)
    control_rep = len(control_list)
    start1 = 0
    end1 = start2 = test_rep
    end2 = test_rep+control_rep
    # if swap 
    if swap is True: 
        test_rep, control_rep = control_rep, test_rep
        #test_list, control_list = control_list, test_list

    # initialize basic array structures
    sig_peaks_list = []    
    
    # single-core version. 
    if parameter.num_procs <2:
        for chr in parameter.chr_info:
            read_array = read_dict[chr]
            sig_peaks_list.extend(per_chr_nbtest(read_array, chr, swap,threshold, peaktype,parameter, start1,end1,start2,end2,test_rep,control_rep))
    # multi-core version
    else: 
        result_list = []
        def log_result(result):
            result_list.append(result)
        try: 
            import sharedmem
            for chr in parameter.chr_info:
                read_array = read_dict[chr]
                read_dict[chr] = sharedmem.copy(read_array)
        except ImportError:
            print("Import sharedmem package failed")
            
        pool = multiprocessing.Pool(processes=parameter.num_procs)#,maxtasksperchild=1)
        for chr in parameter.chr_info:
            read_array = read_dict[chr]
            pool.apply_async(per_chr_nbtest, (read_array, chr, swap,threshold, peaktype, parameter.difftest, 
                       start1,end1,start2,end2,test_rep,control_rep),callback=log_result)
        pool.close()
        pool.join()
        sig_peaks_list = list(itertools.chain(*result_list))
            
    #calculate the BH FDR. 
    debug("begin estimating FDR")
    sig_peaks_list = cal_FDR(sig_peaks_list, num_tests)
    debug("finished estimating FDR")

    # merge adjacent significant peaks. 
    info ("Merging adjacent significant windows...")
    final_peak_list = []
    for chr in read_dict:
        sig_peak_list_by_chr = \
                [item for item in sig_peaks_list if item.chr==chr]
        if len(sig_peak_list_by_chr) == 0:
            continue  # if there is no significant peak in this chromosome, skip it.
        sig_index = [item.index for item in sig_peak_list_by_chr]
        sig_pval = [item.pvalue for item in sig_peak_list_by_chr]
        sig_qval = [item.qvalue for item in sig_peak_list_by_chr]
        sig_g1_count = [item.g1_count for item in sig_peak_list_by_chr]
        #print sig_g1_count, len(sig_g1_count)
        sig_g2_count = [item.g2_count for item in sig_peak_list_by_chr]
        sig_start, sig_end, sig_fc, sig_pval, sig_qval = merge_sig_window(sig_index,
                sig_g1_count, sig_g2_count, sig_pval, sig_qval, peaktype)
        for idx in range(len(sig_start)):
            final_peak = [chr, sig_start[idx]*windowsize/2, 
                          sig_end[idx]*windowsize/2+windowsize, sig_fc[idx],
                          sig_pval[idx], sig_qval[idx]]
            final_peak_list.append(final_peak) 
    # sort the peak list
    final_peak_list = sorted(final_peak_list, key=itemgetter(4))
    info("%d peaks called.", len(final_peak_list))
    if len(final_peak_list) == 0:
        return 
    #start output peaks. 
    all_fc = [peak[3] for peak in final_peak_list] 
    #print all_fc
    max_fc = max(all_fc)
    # write results to peak file.
    peakfile = open(peakfilename, 'w')
    for idx, final_peak in enumerate(final_peak_list): 
        chr = final_peak[0]
        start = final_peak[1]
        end = final_peak[2]
        fc = final_peak[3]
        pval = final_peak[4]
        qval = final_peak[5]
        # tentatively, assign the normalized fold change as the score in 5th column
        score = fc/max_fc*1000 # range from 0 to 1000
        peakfile.write( '\t'.join([chr, str(int(start)), str(int(end)), 
                        ("chip2" if swap else "chip1") + "_peak_" +str(idx+1), 
                        str(score), '.', str(fc), str(pval), str(qval)]) + '\n')
    return 

def merge_sig_window(index_list, g1_count_list, g2_count_list, pval_list, qval_list, peaktype):
    ''' Merge significant windows that are nearby. '''
    if peaktype=="sharp":
        # Mininal number of windows required. 
        MIN_WINDOW = 1 
        # The maximal gap (max-1) of significant windows allowed to merge.
        MAX_WINDOW = 2 
    elif peaktype =="broad":
        MIN_WINDOW = 1 
        MAX_WINDOW = 5
    sig_peak_start = []
    sig_peak_end = []
    sig_fc = []
    sig_pval = []
    sig_qval = []
    for idx, pos in enumerate(index_list):
        if idx == 0:
            sig_peak_start = []
            sig_peak_end = []
            start = pos
            pre_start = pos
            peak_pval_list = [pval_list[idx]]
            peak_qval_list = [qval_list[idx]]
            g1_count = g1_count_list[idx]
            g2_count = g2_count_list[idx]
        else: 
            if pos - pre_start <= MAX_WINDOW:
                pre_start = pos
                peak_pval_list.append(pval_list[idx])
                peak_qval_list.append(qval_list[idx])
                g1_count += g1_count_list[idx]
                g2_count += g2_count_list[idx]
            elif pos-pre_start > MAX_WINDOW:
                end = pre_start
                if (end-start>=MIN_WINDOW-1):
                    sig_peak_start.append(start)
                    sig_peak_end.append(pre_start)
                    sig_pval.append(min(peak_pval_list))
                    sig_qval.append(min(peak_qval_list))
                    sig_fc.append(g1_count/g2_count)
                start = pos
                pre_start = pos
                peak_pval_list = [pval_list[idx]]
                peak_qval_list = [qval_list[idx]]
                g1_count = g1_count_list[idx]
                g2_count = g2_count_list[idx]
                
    if (pre_start-start>MIN_WINDOW-1):
        sig_peak_start.append(start)
        sig_peak_end.append(pre_start)
        sig_pval.append(min(peak_pval_list))
        sig_qval.append(min(peak_qval_list))
        sig_fc.append(g1_count/g2_count)
        
    return sig_peak_start, sig_peak_end, sig_fc, sig_pval, sig_qval




#logConfig.py


import logging
import sys
import time
import os

def startLog(filename):
    #configure root settings.
    logging.basicConfig(level=logging.DEBUG, format = "%(levelname)-6s %(name)s %(asctime)s | %(message)s", datefmt = "%m/%d/%Y %I:%M:%S %p")
    format_basic = logging.Formatter("%(levelname)-7s %(asctime)s | %(message)s", datefmt = "%m/%d/%Y %I:%M:%S %p")

    #Create handlers that prints DEBUG messages to log files
    log_filename = '-'.join([filename, time.strftime("%m-%d-%Y-%I.%M"), str(os.getpid()), "debug.log"])
    # check if the same file name existk, if yes, add an appendix to the file name and then create the file. 
    # although not very likely
   
    file_hand = logging.FileHandler(log_filename, 'w')
    file_hand.setLevel(logging.DEBUG)
    file_hand.setFormatter(format_basic)
    logging.getLogger().addHandler(file_hand)





#!/usr/bin/env python
''' The main .py file for the PePr pipeline '''

# basic modules
import re
import os
import sys
import time
from logging import info

# local modules
from . import optParser
from .pre_processing import initialize
from . import prepareData
from . import sigTests


def pre_processing_module():
    opt = optParser.opt_parser(sys.argv)
    parameter = optParser.process_opt(opt)
    # read data and estimate the shiftsize, normalization constant
    initialize.preprocess(parameter)
    parameter.write_parameter_to_file()

def argless_main():
    '''setuptools entry_points take no arguments, wrap main instead'''
    main(sys.argv)

def main(argv):
    '''PePr main function. Including preprocing and peak calling. 
    No post-processing'''

    opt = optParser.opt_parser(argv)
    parameter = optParser.process_opt(opt)
    # read data and estimate the shiftsize, normalization constant
    initialize.preprocess(parameter)
    parameter.write_parameter_to_file()

    # read data again, begin to process. 
    
    prepareData.read_files_to_arrays(parameter)
    read_dict = prepareData.prepare_data(parameter)
    
    if parameter.difftest is False: 
        swap = False
        peakfilename = parameter.output_directory + parameter.name +"__PePr_peaks.bed"
        sigTests.negative_binomial(read_dict, peakfilename, swap, parameter)
    else: 
        
        up_peakfilename = parameter.output_directory + parameter.name+"__PePr_chip1_peaks.bed"
        swap = False
        sigTests.negative_binomial(read_dict, up_peakfilename, swap, parameter)
        down_peakfilename = parameter.output_directory + parameter.name+"__PePr_chip2_peaks.bed"
        swap = True
        sigTests.negative_binomial(read_dict, down_peakfilename,
                                   swap, parameter)

    info("PePr finished running, thanks for all the wait!")
	
if __name__ == "__main__":
    try:
        main(sys.argv)
    except KeyboardInterrupt:
        print ("user interrupt me")



import os
from logging import info, debug, warning
import collections
import pysam

from . import logConfig
## global variable read_dict
read_dict = {}
def init_dict():
    global read_dict
    read_dict = {}

       
class Parameters:
    "store the parameters that will be used in many processes."
    # ---- input parameters ---- #
    # ---- function specific parameters ---- #

    def __init__(self, opt):
        self.chip1 = []
        self.chip2 = []
        self.input1 = []
        self.input2 = []
        self.file_format = None
        self.shift_size = -1
        self.window_size = -1
        self.difftest = False 
        self.name = None
        self.threshold = None
        self.peaktype = None
        self.normalization = '' 
        self.input_directory = ''
        self.output_directory = ''
        self.keep_max_dup = -1
 
        ## dictionaries storing file related structures. 
        self.read_length_dict = {} # store the read length
        self.shift_dict = {} #store the shift sizes
        self.bin_dict = {} #store the binned reads for normalization.
        self.lib_size_dict = {} # store the library size for normalization.
        self.normalization_dict = {} # store the normalization constants.
        self.array_dict = {} # store the processed read array for significance testing. 
        
        self.num_procs = 1
        # reads specific parameters. 
        self.read_length = -1
        # deprecated parameters. 
        self.remove_artefacts = False
        self.narrow_peak_width = False 
        
        if opt.parameter is not "":
            self.process_parameter_file(opt.parameter)
        else:
            self.process_command_line_option(opt)
        self.validate_parameters()
        self.validate_files()
        logConfig.startLog(self.output_directory + self.name)
        self.check_file_formats()
        # --- initialize logging --- #

    def process_parameter_file(self, parameter_file):
        # using a case/switch commands
        file = open(parameter_file, 'r')
        for line in file:
            items = line.split('#')[0].split()
            # skip empty lines and lines starting with #
            if len(items) == 0:
                continue
            if len(items) > 4:
                raise Exception("There are more than 4 columns in your parameter file.")
            key = items[0].strip().lower()
            value = items[1:]
            
            if key == "chip1":
                self.chip1.append(value[0])
            elif key == "chip2":
                self.chip2.append(value[0])
            elif key == "input1":
                self.input1.append(value[0])
            elif key == "input2":
                self.input2.append(value[0])
            # read the shift size and normalization constants. 
            if key in ['chip1','chip2','input1','input2']:
                if len(value) > 1: 
                    self.shift_dict[value[0]] = int(value[1])
                if len(value) > 2:
                    self.normalization_dict[value[0]] = float(value[2])
                    
            elif key == "file-format":
                self.file_format = value[0].lower()
            elif key == "shiftsize":
                self.shift_size = int(value[0])
            elif key == "windowsize":
                self.window_size = int(value[0])
            elif key == "peaktype":
                self.peaktype = value[0].lower()
            elif key == "difftest":
                self.difftest = value[0].lower() == "true"
            elif key == "name":
                self.name = value[0]
            elif key == "threshold":
                self.threshold = float(value[0])
            elif key == "keep-max-dup":
                self.keep_max_dup = int(value[0])
            elif key == "normalization":
                self.normalization = value[0].lower()
            elif key == "num-processors":
                self.num_procs = int(value[0])
            elif key == "input-directory":
                self.input_directory = value[0]
            elif key == "output-directory":
                self.output_directory = value[0]                
            else:
                raise Exception("Incorrect or unknown parameter: {0}".format(line))
        return 
        
    def process_command_line_option(self, opt):
        self.chip1 = opt.chip1
        self.input1 = opt.input1
        self.chip2 = opt.chip2
        self.input2 = opt.input2
        self.file_format = opt.file_format.lower()
        self.shift_size = opt.shift_size
        self.window_size = opt.window_size
        self.difftest = opt.difftest
        self.name = opt.name
        self.threshold = opt.threshold
        self.peaktype = opt.peaktype.lower()
        self.normalization = opt.normalization
        self.keep_max_dup = opt.keep_max_dup
        self.input_directory = opt.input_directory
        self.output_directory = opt.output_directory
        self.num_procs = opt.num_procs
        return 
        
    def validate_parameters(self):
        # if there are files with the same name, raise an exception. 
        
        # if any of the required files are missing, raise an exception.
        if self.chip1 == []:
            raise Exception("Please specify ChIP-1 sample names")
        if self.input1 == [] and self.difftest is False:
            raise Exception("Please specify input-1 sample names")
        if self.difftest is True:  # also check group 2 files if difftest
            if self.chip2 == []:
                raise Exception("Please specify ChIP-2 sample names")
        # will not test if input2 is available. 
        #    if self.input2 == ['']:
        #        raise Exception("Please specify input-2 sample names")
        if len(self.chip1) < 2 and len(self.input1) <2: 
            raise Exception('''Only 1 replicates detected. To use PePr, at least
                two replicates are required''')
        if self.chip2 != [] and self.input2 != [] and  self.difftest is False: 
            raise Exception('''chip2 and input2 detected, but --diff parameter
                is missing. Please set difftest to True if you want to run differential
                analysis''')
        if not self.file_format:
            raise Exception('''Please specify a file format:  bed,
                sam, bam, sampe or bampe''')
        if self.file_format not in [
                'bed', 'sam', 'bam', 'sampe', 'bampe'
                ]:
            raise Exception('''Please specify a valid file format: bed,
            sam, bam, sampe or bampe''')
        if self.peaktype not in ['sharp', 'broad']:
            raise Exception('''please specify a peak type: sharp or broad.
            Typically, sharp works for TF better and broad
            for histone modifications.''')
       
        ## check normalization
        # if normalization constants not provided
        if  len(self.normalization_dict) == 0:
            if  self.normalization =='':
                if self.difftest is False:
                    self.normalization = "intra-group"
                else: 
                    self.normalization = "inter-group"
            if self.normalization not in ['inter-group','intra-group','scale','no']:
                raise Exception('''Please specify a normalization method: inter-group, intra-group, or scale. put 'no' if you don't want to normalize''')
            
        if self.difftest is True:
            if len(self.chip1) == \
                    len(self.input1):
                self.chip1_matched_input = True
            else:
                self.chip1_matched_input = False
            if len(self.chip2) == \
                    len(self.input2):
                self.chip2_matched_input = True
            else:
                self.chip2_matched_input = False
        
        # add '/' to the end of directories if not. 
        # no '/' at beginning of filenames if input directory is present.
        if self.input_directory != '':
            for filename in self.get_filenames():
                if filename.startswith('/'):
                    raise Exception('If input directory is specified, absolute path("/") in file name is invalid')
            if not self.input_directory.endswith('/'):
                self.input_directory += '/'
        if self.output_directory != '' and \
                not self.output_directory.endswith('/'):
            self.output_directory += '/'

        # can change relative directory to absolute directory. 

    def validate_files(self):
        if self.input_directory != '':
            if os.path.isdir(self.input_directory) is False:
                raise Exception('Input directory: {0} does not exist'.format(self.input_directory))
        if self.output_directory != '':
            if os.path.isdir(self.output_directory) is False:
                raise Exception('Output directory: {0} does not exist'.format(self.output_directory))
        for filename in self.get_filenames():
            if os.path.isfile(self.input_directory+filename) is False:
                print("File:",self.input_directory+filename, " not found")
                exit(1)

    @staticmethod
    def check_sampe_sorted(filename, input_dir):
        with open(input_dir + filename,'r') as infile:
            for line in infile:
                if not line.startswith("@"):
                    break
            count_list = []
            count = 1
            pre_name = ''
            for idx,line in enumerate(infile): 
                name = line.strip().split()[0]
                if name == pre_name:
                    count += 1
                else: 
                    count_list.append(count)
                    count = 1
                    pre_name = name
                if idx == 999:
                    break

        count1 = len([i for i in count_list if i==1])
        ratio = float(count1)/sum(count_list)
        #print filename, ratio
        if ratio > 0.8:
            warning("%s may not be sorted by read name. Please check.",filename)

    @staticmethod
    def check_bampe_sorted(filename, input_dir):
        infile = pysam.Samfile(input_dir + filename, 'rb')
        count_list = []
        count = 1
        pre_name = ''
        for idx,line in enumerate(infile.fetch(until_eof = True)):
            if line.query_name == pre_name:
                count += 1
            else:
                count_list.append(count)
                count = 1
                pre_name = line.query_name
            if idx == 999:
                break

        count1 = len([i for i in count_list if i==1])
        ratio = float(count1)/sum(count_list)
        #print filename, ratio
        if ratio > 0.8:
            warning("%s may not be sorted by read name. Please check.",filename)

    def check_file_formats(self):
        check_name_sorted = {'sampe':self.check_sampe_sorted, 'bampe':self.check_bampe_sorted}
        if self.file_format in ['sampe', 'bampe']:
            for filename in self.get_filenames():
                check_name_sorted[self.file_format](filename, self.input_directory)
        

    def write_parameter_to_file(self):
        '''write the current parameters to the files so user can repeat the analysis'''
        # check if the file name has already be taken. 
        # never mind, will only implement it if really needed.
        output_str = '\t'.join(['#filetype', 'filename', 'shift_size','normalization_factor'])+'\n' 
        
        for filename,filetype in zip(self.get_filenames(), self.get_filetypes()):
            output_str += '\t'.join([filetype,filename,str(self.shift_dict[filename]), str(self.normalization_dict[filename])]) +'\n'
                
        output_str += 'file-format\t'+self.file_format+'\n'
        output_str += 'peaktype\t'+self.peaktype+'\n'
        output_str += 'windowsize\t'+str(self.window_size)+'\n'
        output_str += 'difftest\t'+str(self.difftest)+'\n'
        output_str += 'threshold\t'+str(self.threshold)+'\n'
        output_str += 'normalization\t'+self.normalization+'\n'
        if self.keep_max_dup > 0:
            output_str +='keep-max-dup\t'+str(self.keep_max_dup)+'\n'
        if self.input_directory != '':
            output_str += 'input-directory\t'+self.input_directory+'\n'
        if self.output_directory !='':
            output_str += 'output-directory\t'+self.output_directory+'\n'
        output_str += 'name\t'+self.name+'\n'
        output_str += 'num-processors\t'+str(self.num_procs)+'\n'
        #print and save the parameters into a file. 
        self.print_parameters(output_str)
        with open(self.output_directory + self.name+"__PePr_parameters.txt", 'w') as outfile:
            outfile.write(output_str)
        return 
            
    def print_parameters(self, output_str):
        info("printing running parameters:")
        print('\n\n'+ output_str + '\n')    
        return 
        
    def get_filenames(self):
        return self.chip1+self.chip2+self.input1+self.input2
    
    def get_uniq_filenames(self):
       return list(set(self.get_filenames()))
    
    def get_filetypes(self):
        return ['chip1']*len(self.chip1) + ['chip2'] * len(self.chip2) + ['input1'] * len(self.input1) + ['input2'] * len(self.input2)
    
    def get_filenames_wo_bin_dict(self):
        return [filename for filename in self.get_filenames() if filename not in self.bin_dict ]
    
    def get_chip_filenames(self):
        return self.chip1+self.chip2

    def get_input_filenames(self):
        return self.input1+self.input2
        
    def get_genome_size(self):
        return sum(self.chr_info.values())
    
    def get_top3_chr(self):
        values = [i for i in self.chr_info.values()]
        values.sort()
        values.reverse()
        values_top3 = values[0:3]
        chr_top3 = []
        for value in values_top3:
            for chr in self.chr_info:
                if self.chr_info[chr] == value:
                    chr_top3.append(chr)
        return chr_top3
        
class Peak:
    "data structure that contains the significant peaks"

    def __init__(self, chr, index, group1_count, group2_count, pvalue, qvalue):
        self.chr = chr
        self.index = index
        self.g1_count = group1_count
        self.g2_count = group2_count
        self.pvalue = pvalue
        self.qvalue = qvalue

    def __repr__(self):
        return repr((self.chr, self.index, self.pvalue, self.qvalue))



###### you can use this script to estimate shift size for each sample.
# example: python estimate_shiftSize.py file_format filename
import sys
from shiftSize import estimate_shiftsize
from initialize import get_chromosome_info
from initialize import get_read_length_info


class TestParameter:
    def __init__(self,argv):
        self.file_format = argv[1]
        self.filename = [argv[2]]
        self.chr_info = {} 
        self.input_directory = ''
        self.read_length_dict = {}

    def get_filenames(self):
        return self.filename

if __name__ == "__main__":
    parameter = TestParameter(sys.argv)
    print (sys.argv)
    get_chromosome_info(parameter, sys.argv[2])
    get_read_length_info(parameter)
    print(parameter.chr_info)
    estimate_shiftsize(sys.argv[2], parameter)




#!/usr/bin/env python

import sys
import re
import math
import random
import numpy

from scipy.stats import rankdata
from logging import info 
from .cal_normalization import parse_to_bin

def estimate_window_size(filename, parameter, iter=500):
    '''function to estimate the window size if not provided'''
    
    info ("begin estimating window size")
    bin_size = 20 
    array_window = parse_to_bin(filename, bin_size, parameter)  
    #print array_window
    #print numpy.sum(array_window)
    rowNum = len(array_window)
    peak_len_list = []
    for x in range(iter):
        peak = numpy.max(array_window)
        peak_idx = numpy.where(array_window == peak)[0][0]
        # set the peak to -1 so that it won't confuse the next iteration.
        array_window[peak_idx]= -1 
        if peak_idx == 0:  # Check if the window reaches the left boundary.
            left_boundary_not_reached = False
        else:
            left_boundary_not_reached = True    
        # Check if the window reaches of the right boundary.
        if peak_idx == rowNum-1: 
            right_boundary_not_reached = False
        else: 
            right_boundary_not_reached = True
        i_l = 1
        i_r = 1 
        # Be careful if anything equals to -1. 
        # Though it's unlikely that it will happen.
        while left_boundary_not_reached:

            while (array_window[peak_idx-i_l] >= 0.1*peak): 
                array_window[peak_idx-i_l] = -1
                if peak_idx-i_l==0:
                    left_boundary_not_reached = False
                    break
                i_l += 1
            if left_boundary_not_reached:
                if peak_idx-i_l==0:
                    array_window[peak_idx-i_l] = -1
                    left_boundary_not_reached = False
            
            if left_boundary_not_reached: 
            # Will continue counting if the next window(one window gap) 
            # has above the 10% mode reads. 
                if (array_window[peak_idx-i_l-1] >= 0.1*peak): 
                    array_window[peak_idx-i_l] = -1
                    i_l += 1
                else:
                    break

        while right_boundary_not_reached:
            while (array_window[peak_idx+i_r] >= 0.1*peak):
                array_window[peak_idx+i_r] = -1
                if peak_idx+i_r ==rowNum-1:
                    right_boundary_not_reached = False
                    break
                i_r += 1
            if right_boundary_not_reached:
                if peak_idx+i_r==rowNum-1:
                    array_window[peak_idx+i_r] = -1
                    right_boundary_not_reached = False

            if right_boundary_not_reached:
                if (array_window[peak_idx+i_r+1] >=0.1*peak): 
                    array_window[peak_idx+i_r] = -1
                    i_r += 1
                else:
                    break
        peak_len_list.append((1+i_l+i_r)*bin_size)
    # print peak_len_list
    window_size = int(numpy.median(peak_len_list))
    info ("Estimated window size is %d", window_size)
    if window_size < 100:
        window_size = 100
        info("Estimated window size too narrow, use 100 instead")
    elif window_size > 1000:
        window_size = 1000
        info("Estimated window size too large, use 1000 instead")
    parameter.window_size = window_size
    return 





### module for preprocessing the data to estimate the shift size and normalization constant. 
#from collections import Counter
from logging import info
import pysam

from . import shiftSize
from . import cal_normalization
from . import windowSize

def preprocess(parameter):
    ''' read file and estimate shift sizes and normazliation parameters. 
    '''
    # use one chip sample estimate the chromosome size
    chip_filename = parameter.chip1[0]
    get_chromosome_info(parameter, chip_filename)
    get_read_length_info(parameter)
    if parameter.window_size < 0:
        windowSize.estimate_window_size(chip_filename, parameter)
    # if shift size is not available, estimate the shift size for chip1 and input1
    if parameter.shift_size >= 0:
        for filename in parameter.get_filenames():
                parameter.shift_dict[filename] = parameter.shift_size 
    else:
        # if paired-end reads, set the shift size to 0.
        if parameter.file_format in ['sampe','bampe']:
            for filename in parameter.get_filenames():
                parameter.shift_dict[filename] = 0
        if len(parameter.shift_dict) is 0:
            shiftSize.estimate_shiftsizes(parameter)
    # if normalization constants are not available, estimate it.     
    if len(parameter.normalization_dict) is 0:
        cal_normalization.cal_normalization_constant(parameter)
    
    del parameter.bin_dict 
    return 
    

    
def get_chr_info_bam(parameter, filename):
    chr_info = {}
    num = 0
    infile = pysam.Samfile(parameter.input_directory+filename, 'rb')
    for line in infile.fetch(until_eof=True):
        num += 1
        if num %10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        if line.is_unmapped is False:
            chr = infile.getrname(line.tid)
            try: 
                chr_info[chr] = max(chr_info[chr], line.pos)
            except KeyError: 
                chr_info[chr] = line.pos
    parameter.chr_info = chr_info
    return 
    
def get_chr_info_sam(parameter, filename):
    chr_info = {}
    num = 0
    infile = open(parameter.input_directory+filename, 'r')
    # skip the header of the SAM file. 
    for line in infile:
        if not line.startswith("@"):
            break
    # start reading the real data
    for line in infile:
        num += 1
        if num %10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        words = line.strip().split()
        flag = int(words[1])
    
        if not flag & 0x0004: #if not unmapped
            chr, pos =  words[2], int(words[3])-1
            try: 
                chr_info[chr] = max(chr_info[chr], pos)
            except KeyError: 
                chr_info[chr] = pos
    parameter.chr_info = chr_info
    return  
    
def get_chr_info_bed(parameter, filename):
    chr_info = {}
    infile = open(parameter.input_directory+filename, 'r')
    num = 0
    for line in infile: 
        num += 1
        if num %10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        chr,start,end,col3,col4,strand = line.strip().split()
        try: 
            chr_info[chr] = max(chr_info[chr], int(start))
        except KeyError:
            chr_info[chr] = int(start)
    parameter.chr_info = chr_info
    return


def get_chromosome_info(parameter, chip_filename):
    info ("getting chromosome info")
    if parameter.file_format in ["bam","bampe"]:
         get_chr_info_bam(parameter, chip_filename)   
    if parameter.file_format in ["sam","sampe"]:
         get_chr_info_sam(parameter, chip_filename)        
    if parameter.file_format == "bed":
         get_chr_info_bed(parameter, chip_filename)
    message = ' | '.join([str(key)+":"+str(value) for key,value in parameter.chr_info.items()])
    info("Found "+message)
    return 

def get_read_length_info(parameter):
    if parameter.file_format in ["bam","bampe"]:
        get_read_length_from_bam(parameter)
    if parameter.file_format in ["sam","sampe"]:
        get_read_length_from_sam(parameter)
    if parameter.file_format == "bed":
        get_read_length_from_bed(parameter)
    
    message = '|'.join([ str(x) for x in parameter.read_length_dict.values()])
    info("Read length: "+message)

def get_read_length_from_bam(parameter):
    for filename in parameter.get_filenames():
        with pysam.Samfile(parameter.input_directory+filename, 'rb') as infile:
            length_list = []
            num = 0
            #for idx in range(1000):
            #    line = infile.fetch(until_eof=True).__next__()
            for line in infile.fetch(until_eof=True):
                num += 1
                if num == 1000:
                    break
                try:
                    length_list.append(line.query_length)
                # rlen is deprecated starting from pysam v0.9
                except AttributeError: 
                    length_list.append(line.rlen)
            length = max(length_list)
        parameter.read_length_dict[filename] = length
    return

def get_read_length_from_sam(parameter):
    for filename in parameter.get_filenames():
        with open(parameter.input_directory+filename, 'r') as infile:
            idx = 0
            length_list = []
            for line in infile:
                words = line.strip().split()
                if not words[0].startswith("@"):
                    length_list.append(len(words[9]))
                    idx += 1
                    if idx == 1000:
                        break
            length = max(length_list)
        parameter.read_length_dict[filename] = length
    return 

def get_read_length_from_bed(parameter):
    for filename in parameter.get_filenames():
        with open(parameter.input_directory+filename, 'r') as infile:
            length_list = []
            for idx in range(1000):
                line = infile.readline()
                chr,start,end,col3,col4,strand = line.strip().split()
                length_list.append(int(end)-int(start))
            length = max(length_list)
        parameter.read_length_dict[filename] = length
    return





###############################################################
## script to calcualte the maximum number of duplicates to keep at single position
## usage: python cal_max_duplicates_to_keep.py genome_size library_size
## By Yanxiao Zhang <yanxiazh@umich.edu>
## Timestamp: 5/26/2016
#######################

import sys
from scipy.stats import binom_test

def binomial(n, p):
    #calculate the expected maximum number of replicated reads at a single position
    x = 1
    pvalue = 0
    while (binom_test(x,n,p) > 0.00001):
        x = x + 1
    if x >1:
        x= x - 1
    return x 


def cal_max_dup(argv):
    ''' The main function for calculating the max dup'''
    help = 'usage: python cal_max_duplicates_to_keep.py genome_size library_size'
    if len(argv) != 3:
        print help
        exit(1)
    genome_size = float(argv[1])
    read_total = float(argv[2])
    expected_dup = binomial(read_total, 1.0/genome_size)
    print "The expected maximum duplicate is {0}".format(expected_dup)
    return expected_dup

if __name__ == "__main__":
    cal_max_dup(sys.argv)    



#!/user/bin/env python
# use this script to find the mode of a peak. 
# Input: peak file, read file and format(bed,bam,sam)
# Output: peak file with additional column denoting the peak_mode
### usage: python find_peak_mode.py Peaks_file read_file file_format fragment_size
# Created By Yanxiao Zhang, 5/19/2016
############################################
import sys
import array
import numpy
import pysam 

def parse_sam_for_f_r(filename):
    infile = open(filename, 'r')
    num = 0
    forward = {}
    reverse = {}
    # skip the header of the SAM file.
    for line in infile:
        if not line.startswith("@"):
            break

    words = infile.readline().strip().split()
    length = len(words[9])
    # start reading the real data
    for line in infile:
        num += 1
        if num % 10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        words = line.strip().split()
        flag = int(words[1])

        if not flag & 0x0004: #if not unmapped
            chr, pos =  words[2], int(words[3])-1
            if not flag & 0x0010: # if not reverse
                try:  forward[chr].apend(pos)
                except KeyError:
                    forward[chr] = array.array('i',[pos])
            else:
                try:  reverse[chr].append(pos)
                except KeyError:
                    reverse[chr] = array.array('i',[pos])
    return forward, reverse, readlength

def parse_bam_for_f_r(filename):
    num = 0
    forward = {}
    reverse = {}

    infile =pysam.Samfile(filename, 'rb')
    line = infile.fetch(until_eof=True).__next__()
    length = line.alen
    for line in infile.fetch(until_eof = True):
        num += 1
        if num % 1000000 == 0 :
            print ("{0:,} lines processed in {1}".format(num, filename))
        if line.is_unmapped is False:
            chr = infile.getrname(line.tid)
            if line.is_reverse is False:
                try: forward[chr].append(line.pos)
                except KeyError:
                    forward[chr] = array.array('i',[line.pos])
            else:
                try: reverse[chr].append(line.pos)
                except KeyError:
                    reverse[chr] = array.array('i',[line.pos])
    return forward,reverse, length

def parse_bed_for_f_r(filename):
    infile = open(filename, 'r')
    num = 0
    forward = {}
    reverse = {}
    words = infile.readline().strip().split()
    length = int(words[2]) - int(words[1])
    for line in infile:
        chr,start,end,col3,col4,strand = line.strip().split()
        num += 1
        if num %10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        pos = int(start)
        if strand == "+":
            try: forward[chr].append(pos)
            except KeyError:
                forward[chr] = array.array('i',[pos])
        else:
            try: reverse[chr].append(pos)
            except KeyError:
                reverse[chr] = array.array('i',[pos])

    return forward, reverse, length

def cross_cor(f, r):
    npf = numpy.array(f)
    npr = numpy.array(r)
    cor_list = []
    for x in range(50,302,2):
        #print x
        y = len(numpy.intersect1d(npf+x,npr))
        cor_list.append(y)
    return range(50,302,2)[cor_list.index(max(cor_list))]


def estimate_shift_size(forward,reverse):
    shift_list = []
    print 'estimating shift size:'
    for chr in set(forward.keys())&set(reverse.keys()):
        print chr
        chr_f, chr_r = forward[chr], reverse[chr]
        shift_list.append(cross_cor(chr_f, chr_r))
    shift_size = shift_list[len(shift_list)/2]
    return shift_size/2
    
def shift_reads(forward, reverse,shift):
    print 'shifting reads'
    reads_by_chr = {}
    for chr in set(forward.keys())&set(reverse.keys()):
        chr_f = numpy.array(forward[chr])+shift
        chr_r = numpy.array(reverse[chr])-shift
        reads_by_chr[chr] = numpy.append(chr_f, chr_r)
    return reads_by_chr

def find_peak_mode(peak_filename, reads_by_chr,frag_size):
    out_filename = peak_filename + '.w_mode.txt'
    file_in = open(peak_filename, 'r')
    file_out = open(out_filename, 'w')
    for idx,line in enumerate(file_in):
        if idx > 0 and idx % 1000 == 0:
            print idx, 'peaks processed'
        words = line.strip().split('\t')
        chr = words[0]
        start = int(words[1])
        end = int(words[2])
        others = words[3:]
        window = numpy.zeros(end-start)
        read_chr = reads_by_chr[chr]
        read = read_chr[numpy.where((read_chr > start) & (read_chr < end) )]
        for item in read:
            if item-start-frag_size<0:
                window[0:(item-start+frag_size)]+=1
            else: 
                window[(item-start-frag_size):(item-start+frag_size)] +=1
        mode = numpy.max(window)
        where_mode = numpy.where(window == mode)[0][0]
        where_mode += start
        out_str = '\t'.join(words)+'\t'+str(where_mode)+'\n'
        file_out.write(out_str)

def main(argv):
    ''' usage: python find_peak_mode.py Peaks_file read_file file_format'''
    if len(argv) < 4:
        print '''usage: python find_peak_mode.py Peaks_file read_file file_format'''
        exit(1)

    peak_filename = argv[1]
    read_filename = argv[2]
    file_format = argv[3]
    process_file_for_f_r_dict = {'bam':parse_bam_for_f_r,'sam':parse_sam_for_f_r,'bed':parse_bed_for_f_r}
    forward, reverse,read_length = process_file_for_f_r_dict[file_format](read_filename)
    try: 
        frag_size = int(argv[4])
        shift_size = (frag_size -read_length)/2
    except IndexError:
        shift_size = estimate_shift_size(forward, reverse)
        frag_size = shift_size*2 + read_length
    reads_by_chr = shift_reads(forward,reverse,shift_size)
    find_peak_mode(peak_filename, reads_by_chr,frag_size/2)

if __name__ == "__main__":
    main(sys.argv)



import pysam
import array
import numpy

def parse_bam_for_f_r(filename, input_dir):
    num = 0
    forward = {}
    reverse = {}

    infile =pysam.Samfile(input_dir+filename, 'rb')
    for line in infile.fetch(until_eof = True):
        num += 1
        if num % 10000000 == 0 :
            print ("{0:,} lines processed in {1}".format(num, filename))
        if line.is_unmapped is False:
            chr = infile.getrname(line.tid)
            if line.is_reverse is False:
                try: forward[chr].append(line.reference_start)
                except KeyError:
                    forward[chr] = array.array('i',[line.reference_start])
            else:
                try: reverse[chr].append(line.reference_end)
                except KeyError:
                    reverse[chr] = array.array('i',[line.reference_end])
    return forward,reverse

def parse_sam_for_f_r(filename, input_dir):
    infile = open(input_dir+filename, 'r')
    num = 0
    forward = {}
    reverse = {}
    # skip the header of the SAM file. 
    for line in infile:
        if not line.startswith("@"):
            break
    # start reading the real data
    for line in infile:
        num += 1
        if num % 10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        words = line.strip().split()
        flag = int(words[1])

        if not flag & 0x0004: #if not unmapped
            chr, pos =  words[2], int(words[3])-1
            if not flag & 0x0010: # if not reverse
                try:  forward[chr].append(pos)
                except KeyError:
                    forward[chr] = array.array('i',[pos])
            else:
                pos += len(words[10])
                try:  reverse[chr].append(pos)
                except KeyError:
                    reverse[chr] = array.array('i',[pos])
    return forward, reverse

def parse_bed_for_f_r(filename, input_dir):
    infile = open(input_dir+filename, 'r')
    num = 0
    forward = {}
    reverse = {}
    for line in infile:
        chr,start,end,col3,col4,strand = line.strip().split()
        num += 1
        if num %10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        if strand == "+":
            pos = int(start)
            try: forward[chr].append(pos)
            except KeyError:
                forward[chr] = array.array('i',[pos])
        else:
            pos = int(end)
            try: reverse[chr].append(pos)
            except KeyError:
                reverse[chr] = array.array('i',[pos])

    return forward, reverse

def parse_sampe(filename,input_dir):
    infile = open(input_dir+filename, 'r')
    num = 0
    reads_dict = {}
    flen_dict = {}
    flen_list = array.array('i',[])
    # skip the header of the SAM file.
    for line in infile:
        if not line.startswith("@"):
            pre_name = line.strip().split()[0]
            break
    line_saved = False
    # start reading the real data
    for line in infile:
        num += 1
        if num % 10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        words = line.strip().split()
        name = words[0]
        # if the sequence name has already been processed
        if name == pre_name and line_saved == True:
                continue
        # else
        #initialize the condition
        pre_name = name
        line_saved = False
        # process the new sequence. 
        flag = int(words[1])
        if not flag & 0x0004: #if not unmapped
            chr, pos,flen = words[2], int(words[3])-1, int(words[8])
            if flen != 0: # if the other end mapped to the same chromosome
                try: 
                    reads_dict[chr].append(pos+flen/2)
                    flen_dict[chr].append(flen)
                except KeyError:
                    reads_dict[chr] = array.array('i',[pos+flen/2])
                    flen_dict[chr] = array.array('i', [flen])
                flen_list.append(abs(flen))
                line_saved = True

    # will calculate the median fragment size and remove the reads that have larger fragment size than it. 
    flen_median = numpy.median(flen_list)
    # print flen_median

    for chr in reads_dict:
        flen_chr = numpy.array(flen_dict[chr])
        reads_chr = numpy.array(reads_dict[chr])
        reads_dict[chr] = reads_chr[numpy.where(numpy.abs(flen_chr) <= 2*flen_median)[0]]
        # print chr, len(reads_chr),len(reads_dict[chr])

    return reads_dict 

def parse_bampe(filename, input_dir):
    '''parse paired-end bam file'''

    infile =pysam.Samfile(input_dir+filename, 'rb')
    num = 0
    reads_dict = {}
    flen_dict = {}
    flen_list = array.array('i',[])
    # start proccessing the data.
    line_saved = False
    pre_name = ''
    for line in infile.fetch(until_eof = True):
        num += 1
        if num % 1000000 == 0 :
            print ("{0:,} lines processed in {1}".format(num, filename))
        name = line.query_name
        if name==pre_name and line_saved == True:
            continue
        # else 
        #initialize the condition
        pre_name = name
        line_saved = False
        if line.is_unmapped is False:
            chr = infile.getrname(line.tid)
            # pos = line.pos
            # flen = line.tlen
            if line.tlen !=0:
                try: 
                    reads_dict[chr].append(line.pos+line.tlen/2)
                    flen_dict[chr].append(line.tlen)
                except KeyError:
                    reads_dict[chr] = array.array('i',[line.pos+line.tlen/2])
                    flen_dict[chr] = array.array('i',[line.tlen])
                flen_list.append(abs(line.tlen))
                line_saved = True
    # calculate the median fragment size and remove reads that are two times larger.
    flen_median = numpy.median(flen_list)
    for chr in reads_dict:
        flen_chr = numpy.array(flen_dict[chr])
        reads_chr = numpy.array(reads_dict[chr])
        reads_dict[chr] = reads_chr[numpy.where(numpy.abs(flen_chr) <= 2*flen_median)[0]]
    return reads_dict 

### function dictionary. 
parse_file_by_strand = {'bed':parse_bed_for_f_r,'bam':parse_bam_for_f_r,'sam':parse_sam_for_f_r}
parse_file_pe = {'sampe':parse_sampe, 'bampe':parse_bampe}



#!/usr/bin/env python

###########################################################
### script for removing artifactual peaks in the peak file. 
### Usage: python post_processing_PePr.py --peak peak_file --chip chip.files.sep.by.comma --input input.files.sep.by.comma --file-type type --remove-artefacts --narrow-peak-boundary
### Yanxiao Zhang <yanxiazh@umich.edu> 
### Timestamp 5/24/2016
#######################

import re, os, sys,time
import logging
import numpy
from optparse import OptionParser

from PePr.pre_processing.initialize import get_read_length_info 
from PePr.pre_processing.fileParser import parse_file_by_strand

# optionParser
def opt_parser(argv):
    parser = OptionParser()
    parser.add_option(
        '--peak', action='store',type='string',
        dest='peak', default='',
        help='peak file')
    parser.add_option(
        '--chip', action='store',type='string',
        dest='chip',default='',
        help='chip files separated by comma')
    parser.add_option(
        '--input', action='store', type='string',
        dest='input', default='',
        help='input files separated by comma')
    parser.add_option(
        '--file-type', action='store',type='string',
        dest='type', default='',
        help='read file types. bed, sam, bam')
    parser.add_option(
        '--remove-artefacts', action='store_true',
        dest="remove_artefacts",default=False,
        help="remove peaks that may be caused by excess PCR duplicates")
    parser.add_option(
        '--narrow-peak-boundary',action='store_true',
        dest="narrow_peak",default=False,
        help='make peak width smaller but still contain the core binding region')
    (opt, args)=parser.parse_args(argv)
    if len(argv) == 1:
        parser.print_help()
        exit(1)
    return opt

def process_opt(opt):
    if opt.chip == '':
        raise Exception("Please specify chip samples.")
    if opt.input == '':
        raise Exception("Please specify input samples.")
    if opt.peak == '':
        raise Exception("No peak files.")
    if opt.type =='':
        raise Exception("File type not given.")
    if opt.remove_artefacts == False and opt.narrow_peak == False:
        raise Exception("Please specify at least one post-processing option: --remove-artefacts or --narrow-peak-boundary")
 
    ## start process chip file names.
    opt.chip = opt.chip.strip().split(',')
    opt.input = opt.input.strip().split(',')
    return opt

class ReadData:
    def __init__(self, opt):
        self.chip_filename_list = opt.chip
        self.input_filename_list = opt.input
        self.data_dict_by_strands = {}
        self.read_length_dict = {}
        self.file_format = opt.type
        self.input_directory = ''  

    def get_read_length(self):
        return self.read_length_dict.values()[0]

    def get_filenames(self):
        return self.chip_filename_list + self.input_filename_list


def main(argv):
    ## performing the option parser
    opt = opt_parser(argv)
    opt = process_opt(opt)
    readData = ReadData(opt)
    get_read_length_info(readData)
    readData.read_length = readData.get_read_length()
    print('read length is {0}'.format(readData.read_length))
    data_dict = {}

    for filename in opt.chip+opt.input:
        print ("reading {0}".format(filename))
        data_dict[filename] = {}
        forward, reverse = parse_file_by_strand[readData.file_format](filename,readData.input_directory)
        for chr in set(forward.keys())&set(reverse.keys()):
            data_dict[filename][chr] = {}
            data_dict[filename][chr]['f'] = numpy.array(forward[chr])
            data_dict[filename][chr]['r'] = numpy.array(reverse[chr])
    readData.data_dict_by_strands = data_dict     
    
    file_in = open(opt.peak, 'r')
    file_passed_name = opt.peak
    file_failed_name = opt.peak
    if opt.remove_artefacts == True:
        file_passed_name += '.passed'
        file_failed_name += '.failed'
    if opt.narrow_peak == True:
        file_passed_name += '.boundary_refined'
        file_failed_name += '.boundary_refined'

    file_passed = open(file_passed_name, 'w')
    file_failed = open(file_failed_name, 'w')

    post_processing(readData,file_in,file_passed,file_failed, opt.remove_artefacts, opt.narrow_peak)

def post_processing(readData, peak, passed, filtered, remove_artefacts=False, narrow_peak=False):
    print(" Begin post-processing.")
    chip_list =  readData.chip_filename_list
    input_list = readData.input_filename_list
    strands_dict = readData.data_dict_by_strands
    removed_count = 0
    for idx,line in enumerate(peak):
        if idx%100 == 0:
            print ('{0} peaks processed'.format(idx))
        words = line.strip().split('\t')
        chr = words[0]
        start = int(words[1])
        end = int(words[2])
        rest = words[3:]
        (start, end, chip_input_ratio, overlap_orig,
        overlap_roll) = post_processing_per_peak(
            strands_dict, chip_list, input_list, chr, start, end,
            readData.read_length,
            narrow_peak, remove_artefacts)
        # print start,end, chip_input_ratio, overlap_orig, overlap_roll
        if remove_artefacts and (chip_input_ratio > 0.5 or
                (overlap_orig > 0.2 and
                overlap_roll/overlap_orig < 0.5)):
            filtered.write(chr+"\t" + str(start) + '\t' + str(end) )
            for item in rest: 
                filtered.write( '\t' + item )
            filtered.write('\n')
        #    filtered.write(str(chip_input_ratio) + '\t' +
        #                '\t' + str(overlap_orig) +
        #                '\t' + str(overlap_roll) + '\n')
            removed_count += 1
            continue
        passed.write(chr+"\t" + str(start) + '\t' + str(end) ) 
        for item in rest:
            passed.write('\t' + item )
        passed.write('\n')
        ## will not write additional information
        #passed.write(str(chip_input_ratio) + '\t' +
        #                  str(overlap_orig) +
        #                 '\t' + str(overlap_roll) + '\n')
    print ('Done. {0} out of {1} peaks failed the test and removed'.format(removed_count, idx+1))
    return 


def post_processing_per_peak(strands_dict, chip_list, input_list, chr,
                             start, end, readLength, narrow_peak,
                             remove_artefacts):
    ''' Remove artefacts that may be caused by PCR duplicates and/or refine peak boundary'''

    chip_forward = numpy.zeros(end-start)
    chip_reverse = numpy.zeros(end-start)
    input_forward = numpy.zeros(end-start)
    input_reverse = numpy.zeros(end-start)
    for chip in chip_list:
        forward = strands_dict[chip][chr]['f']
        reverse = strands_dict[chip][chr]['r']
        forward_read = forward[numpy.where( (forward >= start) &
                (forward < end) )]
        reverse_read = reverse[numpy.where( (reverse >= start+readLength) &
                (reverse < end+readLength) )]
        for read in forward_read:
            try: chip_forward[(read-start)] +=1
            except IndexError:
                debug("index error ignored. end-start: %d. read-start: %d.", end-start, read-start)
        for read in reverse_read:
            try: chip_reverse[(read-start-readLength)] +=1
            except IndexError:
                debug("index error ignored. end-start: %d. read-start-readLength: %d.", end-start, read-start-readLength)
    # using input reads to remove artefacts
    if remove_artefacts is True:
        for input in input_list:
            forward = strands_dict[input][chr]['f']
            reverse = strands_dict[input][chr]['r']
            forward_read = forward[numpy.where( (forward >=
                    start) & (forward < end) )]
            reverse_read = reverse[numpy.where( (reverse >=
                    start+readLength) & (reverse < end+readLength) )]
            for read in forward_read:
                try: input_forward[(read-start)] +=1
                except IndexError:
                    debug("index error ignored. end-start: %d. read-start: %d.", end-start, read-start)
            for read in reverse_read:
                try: input_reverse[(read-start-readLength)] +=1
                except IndexError:
                    debug("index error ignored. end-start: %d. read-start-readLength: %d.", end-start, read-start-readLength)

        chip_both = chip_forward + chip_reverse
        input_both = input_forward + input_reverse
        if sum(chip_both) == 0:
            pass
        else:
            chip_both = chip_both/sum(chip_both)
        if sum(input_both) ==0:
            pass
        else:
            input_both = input_both/sum(input_both)
        overlap_chip_input = numpy.sum(numpy.minimum(chip_both, input_both))
        if sum(chip_reverse) != 0:
            chip_reverse = chip_reverse/sum(chip_reverse)
        chip_forward_roll = numpy.roll(chip_forward,readLength)
        if sum(chip_forward) != 0:
            chip_forward = chip_forward/sum(chip_forward)
            chip_forward_roll = chip_forward_roll/sum(chip_forward)
        overlap_orig = numpy.sum(numpy.minimum(chip_forward, chip_reverse))
        overlap_orig = numpy.max([overlap_orig, 1e-5])
        overlap_roll = numpy.sum(numpy.minimum(chip_forward_roll, chip_reverse))
    else:
        overlap_chip_input = 0
        overlap_orig = 1e-5
        overlap_roll = 0
    
    # narrow peak width
    if narrow_peak is True:
        sum_forward = 0
        sum_reverse = 0
        if sum(chip_forward) > 0:
            chip_forward = chip_forward/sum(chip_forward)
            for i in xrange(end-start):
                sum_forward += chip_forward[i]
                if sum_forward > 0.2:
                    new_start = start + i
                    break
        if sum(chip_reverse) > 0:
            chip_reverse = chip_reverse/sum(chip_reverse)
            for i in xrange(end-start-1, -1, -1):
                sum_reverse += chip_reverse[i]
                if sum_reverse > 0.2:
                    new_end = start + i
                    break
        if sum(chip_forward) == 0:
            new_start = start #new_end - 2*shiftSize
        if sum(chip_reverse) == 0:
            new_end = end # new_start + 2*shiftSize
        if new_start+ readLength > new_end:
            # if the peak width is negative, use the original peak boundary. 
            new_start = start
            new_end = end
        start = new_start
        end = new_end

    return (start, end, overlap_chip_input,
            overlap_orig, overlap_roll)


def post_processing_module():
    '''setuptools entry_points'''
    main(sys.argv)

if __name__ == '__main__':
    try: main(sys.argv)
    except KeyboardInterrupt:
        print("user interrupted me")






#!/usr/bin/env python
from math import exp
import logging
from scipy.stats import binom_test

debug = logging.debug


def binomial(n, p):
    #calculate the expected maximum number of replicated reads at a single position

    x = 1
    pvalue = 0
    while (binom_test(x,n,p) > 0.00001):
        x = x + 1
    if x >1:
        x= x - 1
    return x 


def median(list):
    '''Will return the median of the list of numbers '''
    list.sort()
    if len(list) % 2 == 0:
        med = (list[int(len(list)/2)]+list[int(len(list)/2-1)])/2
    else:
        med = list[int((len(list)-1)/2)]
    return med

def erfcc(x):
    """Complementary error function."""
    z = abs(x)
    t = 1. / (1. + 0.5*z)
    r = t * exp(-z*z-1.26551223+t*(1.00002368+t*(.37409196+
    	t*(.09678418+t*(-.18628806+t*(.27886807+
    	t*(-1.13520398+t*(1.48851587+t*(-.82215223+
    	t*.17087277)))))))))
    if (x >= 0.):
    	return r
    else:
    	return 2. - r

def ncdf(x):
    return 1. - 0.5*erfcc(x/(2**0.5))




#from logging import info
import multiprocessing
from scipy.stats import rankdata
from logging import info 
import numpy
import itertools 
import pysam 
import array

from .fileParser import parse_file_by_strand

BIN = 1000

def estimate_shiftsizes(parameter):
    ''' the root function for estimating the shiftsizes'''
    
    ## estimate the shift size for chip samples. 
    if parameter.num_procs < 2:
        for chip in parameter.get_chip_filenames():
            shift_size, bin_array = estimate_shiftsize(chip, parameter)
            parameter.shift_dict[chip] = shift_size
            parameter.bin_dict[chip] = bin_array
    else: 
        pool = multiprocessing.Pool(parameter.num_procs)
        p = pool.map_async(estimate_shiftsize_wrapper, zip(parameter.get_chip_filenames(), itertools.repeat(parameter)),1)
        
        try: results = p.get()
        except KeyboardInterrupt:
            exit(1)
        for chip, result in zip(parameter.get_chip_filenames(), results):
            parameter.shift_dict[chip] = result[0]
            parameter.bin_dict[chip] = result[1]
           
    # put the shift size for input samples
    if len(parameter.input1) > 0:
        for input in parameter.input1: 
            shift_list = [parameter.shift_dict[file] for file in parameter.chip1]
            parameter.shift_dict[input] = sum(shift_list)/len(shift_list)
    if len(parameter.input2) > 0:
        for input in parameter.input2: 
            shift_list = [parameter.shift_dict[file] for file in parameter.chip2]
            parameter.shift_dict[input] = sum(shift_list)/len(shift_list) 
    return 
def estimate_shiftsize_wrapper(args):
    try:
        return estimate_shiftsize(*args)
    except KeyboardInterrupt:
        pass
        
def estimate_shiftsize(chip, parameter):
    ''' estimate the shiftsize for each file '''
    info("estimating the shift size for %s", chip)
    info_dict = {} # saving the info matrix for deriving the shift size data. 
    bin_dict = {}
    
    for chr in parameter.chr_info:
        row_num = int(parameter.chr_info[chr]/BIN) 
        bin_dict[chr] = numpy.zeros(row_num, dtype=numpy.float64)
        info_dict[chr] = numpy.zeros((row_num,4),dtype=numpy.int64)
    # parsing file into strand 
    forward, reverse = parse_file_by_strand[parameter.file_format](chip, parameter.input_directory)
    shift_list =[]
    for chr in parameter.get_top3_chr():
        chr_f,chr_r = forward[chr],reverse[chr]
        shift_list.append(cross_cor(chr_f,chr_r))
    shift_list.sort()
    frag_size = shift_list[int(len(shift_list)/2)]
    ### parse the reads into bins
    for chr in parameter.chr_info:
        if chr in forward:
            for pos in forward[chr]:
                try: bin_dict[chr][int(pos/BIN)] += 1 
                except IndexError:
                #    print("Index Error")
                    pass
        if chr in reverse:
            for pos in reverse[chr]:
                try: bin_dict[chr][int(pos/BIN)] += 1
                except IndexError:
                #    print("Index Error")
                    pass
        try: bin_array = numpy.append(bin_array, bin_dict[chr])
        except UnboundLocalError:
            bin_array = bin_dict[chr]

    #frag_size += parameter.read_length_dict[chip]
    shift_size = int(frag_size/2)
    info("The shift size for %s is %d", chip, shift_size)
    return (shift_size, bin_array)

def cross_cor(f, r):
    npf = numpy.array(f)
    npr = numpy.array(r)
    cor_list = []
    for x in range(50,302,2):
        #print x
        y = len(numpy.intersect1d(npf+x,npr))
        cor_list.append(y)
    return range(50,302,2)[cor_list.index(max(cor_list))]
 



import numpy
from logging import info
import pysam
import multiprocessing
import itertools 
#from pre_processing.shiftSize import parse_bed_for_f_r,parse_bam_for_f_r,parse_sam_for_f_r
from .pre_processing.fileParser import parse_file_by_strand, parse_file_pe

def remove_duplicate(fragments, max):
    for chr in fragments:
        fragment_chr = fragments[chr]
        fragment_chr.sort()
        idx_keep = numpy.zeros(fragment_chr.size)
        pos_pre = 0
        count = 1
        for idx,pos in enumerate(fragment_chr):
            if pos == pos_pre:
                count += 1
                if count > max:
                    idx_keep[idx] = count 
            else:
                count = 1
                pos_pre = pos
        fragments[chr] = fragment_chr[numpy.where(idx_keep==0)]
    return fragments 
        

def read_file_to_array(filename, parameter):
    ''' read file into arrays that can be used on significance testing. '''
    info ("processing %s", filename)
    move_size = parameter.window_size/2
    data_dict = {}
    for chr in parameter.chr_info:
        row_num = int(parameter.chr_info[chr]/move_size) - 1
        if row_num < 1: # if the chromosome length is less than window size. 
            row_num = 1 
        data_dict[chr] = numpy.zeros(row_num, dtype=numpy.float64)
    if parameter.file_format.endswith('pe'):
        fragments_list = [parse_file_pe[parameter.file_format](filename,parameter.input_directory)]
    else:
        forward, reverse = parse_file_by_strand[parameter.file_format](filename,parameter.input_directory)
        for chr in forward:
            forward[chr] = numpy.array(forward[chr]) + parameter.shift_dict[filename] 
        for chr in reverse:
            reverse[chr] = numpy.array(reverse[chr]) - parameter.shift_dict[filename] 
        fragments_list = [forward, reverse]
 
    for fragments in fragments_list: 
        if parameter.keep_max_dup > 0:
            fragments = remove_duplicate(fragments,parameter.keep_max_dup)
        for chr in fragments:
            for pos in fragments[chr]:
                try:
                    data_dict[chr][int(pos/move_size)] += 1
                # index out of range at the end of chr.
                except (IndexError, KeyError) as e: pass 
        
    for chr in data_dict: 
        data_dict[chr] = data_dict[chr] + numpy.roll(data_dict[chr],-1)
        data_dict[chr] = data_dict[chr] * parameter.normalization_dict[filename]
    return data_dict 
     
def read_file_to_array_wrapper(args):
    try: 
        return read_file_to_array(*args)
    except KeyboardInterrupt:
        pass
     
def read_files_to_arrays(parameter):
    
    if parameter.num_procs <2:
        for filename in parameter.get_uniq_filenames():
            parameter.array_dict[filename] = read_file_to_array(filename, parameter)
    else:
        pool = multiprocessing.Pool(parameter.num_procs)
#        p = pool.map_async(read_file_to_array_wrapper, zip(parameter.get_uniq_filenames(), itertools.repeat(parameter)),1)
        p = pool.map_async(read_file_to_array_wrapper, zip(parameter.get_uniq_filenames(), itertools.repeat(parameter)),1)
        try: results = p.get()
        except KeyboardInterrupt:
            exit(1)
        for filename, result in zip(parameter.get_uniq_filenames(), results):
            parameter.array_dict[filename] = result
            
    return 
        
def prepare_data_peak_calling(parameter):
    info ("Begin peak-calling")
    data_dict = {}
    files_remain_in_queue = parameter.get_filenames()
    for filename in parameter.chip1:
        for chr in parameter.chr_info:
            try: 
                data_dict[chr] = numpy.column_stack((data_dict[chr], parameter.array_dict[filename][chr]))
            except KeyError:
                data_dict[chr] = parameter.array_dict[filename][chr]
            if filename not in files_remain_in_queue: 
                del parameter.array_dict[filename][chr] 
        files_remain_in_queue.remove(filename)
    for filename in parameter.input1:
        for chr in parameter.chr_info:
            data_dict[chr] = numpy.column_stack((data_dict[chr], parameter.array_dict[filename][chr]))
            if filename not in files_remain_in_queue:
                del parameter.array_dict[filename][chr]        
        files_remain_in_queue.remove(filename)
    return data_dict
    
    
def prepare_data_diff_binding(parameter):
    info ("Begin differential binding analysis")
    data_dict = {}
    files_remain_in_queue = parameter.get_filenames()
    chip1_array = {}
    for filename in parameter.chip1:
        for chr in parameter.chr_info:
            try: 
                chip1_array[chr] = numpy.column_stack((chip1_array[chr], parameter.array_dict[filename][chr]))
            except KeyError:
                chip1_array[chr] = parameter.array_dict[filename][chr]
            # delete the original array. 
            if filename not in files_remain_in_queue:
                del parameter.array_dict[filename][chr]
        files_remain_in_queue.remove(filename)
    if len(parameter.input1)> 0:
        input1_array = {}
        if parameter.chip1_matched_input is True: 
            for filename in parameter.input1:
                for chr in parameter.chr_info:
                    try: 
                        input1_array[chr] = numpy.column_stack((input1_array[chr], parameter.array_dict[filename][chr]))
                    except KeyError:
                        input1_array[chr] = parameter.array_dict[filename][chr]     
                    if filename not in files_remain_in_queue:
                        del parameter.array_dict[filename][chr]
                files_remain_in_queue.remove(filename)
        else: 
            for filename in parameter.input1:
                for chr in parameter.chr_info:
                    try: 
                        input1_array[chr] += parameter.array_dict[filename][chr]
                    except KeyError:
                        input1_array[chr] = parameter.array_dict[filename][chr] 
                    if filename not in files_remain_in_queue:
                        del parameter.array_dict[filename][chr]
                files_remain_in_queue.remove(filename)

            for chr in parameter.chr_info:
                input1_array[chr] /= len(parameter.input1)
                input1_array[chr].resize(len(input1_array[chr]),1)
        for chr in parameter.chr_info:        
            chip1_array[chr] -= input1_array[chr]
            chip1_array[chr][chip1_array[chr]<0] = 0 
    
    # do the same thing for chip2 
    chip2_array = {}
    for filename in parameter.chip2:
        for chr in parameter.chr_info:
            try: 
                chip2_array[chr] = numpy.column_stack((chip2_array[chr], parameter.array_dict[filename][chr]))
            except KeyError:
                chip2_array[chr] = parameter.array_dict[filename][chr]
            if filename not in files_remain_in_queue:
                del parameter.array_dict[filename][chr]    
        files_remain_in_queue.remove(filename)
    if len(parameter.input2)> 0:
        input2_array = {}
        if parameter.chip2_matched_input is True: 
            for filename in parameter.input2:
                for chr in parameter.chr_info:
                    try: 
                        input2_array[chr] = numpy.column_stack((input2_array[chr], parameter.array_dict[filename][chr]))
                    except KeyError:
                        input2_array[chr] = parameter.array_dict[filename][chr]
                    if filename not in files_remain_in_queue:
                        del parameter.array_dict[filename][chr]
                files_remain_in_queue.remove(filename)
        else: 
            for filename in parameter.input2:
                for chr in parameter.chr_info:
                    try: 
                        input2_array[chr] += parameter.array_dict[filename][chr]
                    except KeyError:
                        input2_array[chr] = parameter.array_dict[filename][chr] 
                    if filename not in files_remain_in_queue:
                        del parameter.array_dict[filename][chr]
                files_remain_in_queue.remove(filename)
            for chr in parameter.chr_info:
                input2_array[chr] /= len(parameter.input2)
                input2_array[chr].resize(len(input2_array[chr]),1)
        for chr in parameter.chr_info:
            chip2_array[chr] -= input2_array[chr]
            chip2_array[chr][chip2_array[chr]<0] = 0    
        
    for chr in parameter.chr_info:
        data_dict[chr] =numpy.column_stack((chip1_array[chr], chip2_array[chr]))
        del chip1_array[chr]
        del chip2_array[chr]
    return data_dict
    
    
def prepare_data(parameter):
    ''' The wrapper function to prepare data. 
    Arrange the data into arrays for significance testing. '''
    if parameter.difftest == False:
        read_dict = prepare_data_peak_calling(parameter)
    else: 
        read_dict = prepare_data_diff_binding(parameter)
    # remove the array dict when it is read. 
    del parameter.array_dict 
    for chr in parameter.chr_info:
        read_array = read_dict[chr]
        read_array[numpy.where(read_array ==0)] = 1
    return read_dict
        
        
        



#from distutils.core import setup
from setuptools import find_packages, setup
from PePr import __version__
setup(name="PePr",
      version=__version__, # change the version info in the PePr.__init__ file
      author="Yanxiao Zhang",
      author_email="troublezhang@gmail.com",
      url="https://github.com/shawnzhangyx/PePr/",
      license="GNU GPL v3",
      description="Peak-calling and Prioritization pipeline for replicated ChIP-Seq data",
      long_description="Peak-calling and Prioritization pipeline for replicated ChIP-Seq data",
      platforms = ['any'],
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "Intended Audience :: Science/Research",
          "Topic :: Scientific/Engineering :: Bio-Informatics",
          "Programming Language :: Python",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
          ],
      packages=find_packages(),
      package_data={"PePr": ['data/*.bed']},
      install_requires=[
          'numpy>=1.6.0',
          'scipy>=0.14.0',
          'pysam',
          'sharedmem',
          ],
      entry_points={
          'console_scripts': [
              'PePr=PePr.PePr:argless_main',
              'PePr-preprocess=PePr.PePr:pre_processing_module',
              'PePr-postprocess=PePr.post_processing.post_process_PePr:post_processing_module',
              ]
          }
      )



from logging import info
from logging import debug
import array
import pysam 
import numpy
from scipy.stats import rankdata
import multiprocessing
import itertools


BIN = 1000 # bin size for normalization purposes. 

def cal_normalization_constant(parameter):
    info ("calculating normalization constants")
    if parameter.normalization == "scale":
        scale(parameter)
    elif parameter.normalization == "intra-group":
        within_group_tmm(parameter)
    elif parameter.normalization == "inter-group":
        between_group_tmm(parameter)
    else: #no normalization
        no_normalization(parameter)

def no_normalization(parameter):
    for filename in parameter.get_filenames():
        parameter.normalization_dict[filename] = 1.0
    return

def scale(parameter):
    ''' scale the libaries so the total number of reads are the same'''

    if parameter.num_procs < 2:
        for filename in parameter.get_filenames():
            #print filename
            line_num = get_file_line_counts(filename, parameter)
            parameter.lib_size_dict[filename] = line_num
    else:
        pool =  multiprocessing.Pool(parameter.num_procs)
        p = pool.map_async(get_file_line_counts_wrapper, zip(parameter.get_filenames(), itertools.repeat(parameter)),1)
        try: results = p.get()
        except KeyboardInterrupt:
            exit(1)
        for filename, line_num in zip(parameter.get_filenames(), results):
            parameter.lib_size_dict[filename] = line_num

    lib_size_mean = sum(parameter.lib_size_dict.values())/len(parameter.lib_size_dict)
    for filename in parameter.get_filenames():
        parameter.normalization_dict[filename] = float(lib_size_mean)/parameter.lib_size_dict[filename]     
    return 
        
def get_file_line_counts_wrapper(args):
    try:
        return get_file_line_counts(*args)
    except KeyboardInterrupt:
        pass

def get_file_line_counts(filename, parameter):
    # keep the line count. 
    idx = 0
    #print filename
    if parameter.file_format == "bed":
        with open(parameter.input_directory + filename, 'r') as infile: 
            for idx, line in enumerate(infile):
                pass
            idx += 1    
    elif parameter.file_format in ["sam","sampe"]:
        with open(parameter.input_directory + filename, 'r') as infile:
            for line in infile:
                if not line.startswith('@'): #skip the header lines.
                    break
            for line in infile:
                words = line.strip().split()
                flag = int(words[1])
                if not flag & 0x004:
                    idx += 1
    elif parameter.file_format in ["bam","bampe"]:
        with pysam.Samfile(parameter.input_directory + filename, 'rb') as infile:
            for line in infile.fetch(until_eof = True):
                if line.is_unmapped is False:
                    idx += 1
                    
    return idx
        

### within-group normalization.  
def within_group_tmm(parameter):
    '''within_group normalization '''
    # process the reads into bins. 
    bin_dict = parameter.bin_dict 
    bin_size = BIN
    
    if parameter.num_procs < 2: 
        for filename in parameter.get_filenames_wo_bin_dict():
            bin_dict[filename] = parse_to_bin(filename, bin_size, parameter)
    
    else: 
        pool = multiprocessing.Pool(parameter.num_procs)
        p = pool.map_async(parse_to_bin_wrapper, zip(parameter.get_filenames_wo_bin_dict(), itertools.repeat(bin_size), itertools.repeat(parameter)),1)
        try: results = p.get()
        except KeyboardInterrupt:
            exit(1)
            
        for filename, array in zip(parameter.get_filenames_wo_bin_dict(), results):
            parameter.bin_dict[filename] = array
            
    # process chip1 array 
    for chip in parameter.chip1:
        try: 
            chip1_array_mixed = chip1_array_mixed + bin_dict[chip] 
            chip1_rep_rank_sum += rankdata(-bin_dict[chip])
        except UnboundLocalError:
            chip1_array_mixed = bin_dict[chip].copy()
            chip1_rep_rank_sum = rankdata(-bin_dict[chip])
            
    chip1_array_mixed /= len(parameter.chip1)
    
    # process chip2 array if there is any
    if len(parameter.chip2) > 0:
        for chip in parameter.chip2:
            try: 
                chip2_array_mixed = chip2_array_mixed + bin_dict[chip] 
                chip2_rep_rank_sum += rankdata(-bin_dict[chip])
            except UnboundLocalError:
                chip2_array_mixed = bin_dict[chip].copy()
                chip2_rep_rank_sum = rankdata(-bin_dict[chip])
                
        chip2_array_mixed /= len(parameter.chip2)
        # will scale chip1 and chip2 if their lib sizes are very different. 
        chip1_size = numpy.sum(chip1_array_mixed) 
        chip2_size = numpy.sum(chip2_array_mixed)
        chip1_array_mixed *= (chip1_size+chip2_size)/chip1_size/2
        chip2_array_mixed *= (chip1_size+chip2_size)/chip2_size/2
        
    for chip in parameter.chip1: 
        parameter.normalization_dict[chip] = chip_tmm(chip1_array_mixed, bin_dict[chip], chip1_rep_rank_sum)
    
    if len(parameter.chip2) > 0:
        for chip in parameter.chip2:
            parameter.normalization_dict[chip] = chip_tmm(chip2_array_mixed, bin_dict[chip], chip2_rep_rank_sum)
    if len(parameter.input1) > 0:
        for input in parameter.input1:
            parameter.normalization_dict[input] = input_ncis(chip1_array_mixed, bin_dict[input])
    if len(parameter.input2) > 0:
        for input in parameter.input2:
            parameter.normalization_dict[input] = input_ncis(chip2_array_mixed, bin_dict[input])

    return

def between_group_tmm(parameter):
    '''between group tmm method'''
    # process the reads into bins.
    bin_dict = parameter.bin_dict
    bin_size = BIN

    if parameter.num_procs < 2:
        for filename in parameter.get_filenames_wo_bin_dict():
            bin_dict[filename] = parse_to_bin(filename, bin_size, parameter)
    else:
        pool = multiprocessing.Pool(parameter.num_procs)
        p = pool.map_async(parse_to_bin_wrapper, zip(parameter.get_filenames_wo_bin_dict(), itertools.repeat(bin_size), itertools.repeat(parameter)),1)
        try: results = p.get()
        except KeyboardInterrupt:
            exit(1)

        for filename, array in zip(parameter.get_filenames_wo_bin_dict(), results):
            parameter.bin_dict[filename] = array

    # process chip array
    for chip in parameter.get_chip_filenames():
        try:
            chip_array_mixed = chip_array_mixed + bin_dict[chip]
            chip_rep_rank_sum += rankdata(-bin_dict[chip])
        except UnboundLocalError:
            chip_array_mixed = bin_dict[chip].copy()
            chip_rep_rank_sum = rankdata(-bin_dict[chip])

    chip_array_mixed /= len(parameter.get_chip_filenames())

    for chip in parameter.get_chip_filenames():
        parameter.normalization_dict[chip] = chip_tmm(chip_array_mixed, bin_dict[chip], chip_rep_rank_sum)

    if len(parameter.get_input_filenames()) > 0:
        for input in parameter.get_input_filenames():
            parameter.normalization_dict[input] = input_ncis(chip_array_mixed, bin_dict[input])

    return

    
    
def chip_tmm(ref, target, rep_rank_sum): 
    '''Estimate the ChIP normalization constant against the mixed ChIP
       using the modified TMM method'''
    N_PEAKS_GRID = numpy.array([1000, 5000, 10000, 20000, 30000, 40000, 50000])
    TRIM_M = 0.2
    TRIM_A = 0.05
    order = numpy.argsort(rep_rank_sum)
#    order = numpy.argsort(-ref)
    len_target_not_zero = len(numpy.where(target > 0)[0])
    #print len_target_not_zero
    ref = ref[order]
    target = target[order]
    tmm_array = numpy.array([])
    for idx,n in enumerate(N_PEAKS_GRID):
        if n > len_target_not_zero:
            if idx == 0:
                # if the non-zero list is smaller than 1000
                n = len_target_not_zero
            else:
                break
        ref_n = ref[range(n)]
        ref_n[ref_n==0] = 1
        target_n = target[range(n)]
        target_n[target_n==0] = 1
        Mg = numpy.log2(ref_n/target_n)
        Ag = 0.5*numpy.log2(ref_n*target_n)
        Mg_sorted = Mg.copy()
        Mg_sorted.sort()
        Mg_lower_bound = Mg_sorted[int(n*TRIM_M)]
        Mg_upper_bound = Mg_sorted[int(n*(1-TRIM_M))]
        Ag_sorted = Ag.copy()
        Ag_sorted.sort()
        Ag_lower_bound = Ag_sorted[int(n*TRIM_A)]
        Ag_upper_bound = Ag_sorted[int(n*(1-TRIM_A))]
        trim_index = numpy.where((Mg > Mg_lower_bound) &
            (Mg < Mg_upper_bound) & (Ag > Ag_lower_bound) &
            (Ag < Ag_upper_bound))
        ref_trim = ref_n[trim_index]
        target_trim = target_n[trim_index]
        Mgk = numpy.log2(ref_trim/target_trim)
        Wgk = 0.5*numpy.log2(ref_trim*target_trim)
        tmm = 2**(numpy.sum(Mgk*Wgk)/numpy.sum(Wgk))
        tmm_array = numpy.append(tmm_array, tmm)
        #debug("The TMM estiamted from top %s windows is %s", n, tmm)
    library_ratio = numpy.sum(ref)/numpy.sum(target)
    tmm_diff_array = numpy.abs(tmm_array - library_ratio)
    # print tmm_array, tmm_diff_array
    tmm_max = tmm_array[numpy.argmax(tmm_diff_array)]
    return tmm_max

    
def input_ncis(ref, target):
    '''Estimate the input normalization constant using the NCIS method'''
    combined = ref + target 
    READ_MAX = 200
    MIN_GENOME_COVERAGE = 0.75
    pre_ratio = 1
    #file = open(str(sum(target))+"input_norm.txt",'w')
    for r_cut in range(1, READ_MAX):
        index = numpy.where(combined <= r_cut)[0]
        percent_genome_covered = float(len(index))/len(combined)
        target_sum = numpy.sum(target[index]) or 1.0
        ratio = numpy.sum(ref[index])/target_sum
        #file.write(str(r_cut)+'\t'+str(numpy.sum(ref[index]))+ '\t' +
        #        str(target_sum) +'\t'+ str(ratio)+ '\n')
        if percent_genome_covered > MIN_GENOME_COVERAGE and ratio > pre_ratio:
            return ratio
        else: 
            pre_ratio = ratio
    #file.close()
    return ratio
    
def parse_to_bin_wrapper(args):
    try: 
        return parse_to_bin(*args)
    except KeyboardInterrupt:
        pass 
        
    
    
def parse_bed_to_bin(filename, bin_size, bin_dict, input_dir):
    ''' parse the bed files into bin '''
    infile = open(input_dir + filename, 'r')
    num = 0
    for line in infile: 
        num += 1
        if num %10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        chr,start,end,col3,col4,strand = line.strip().split()
        pos = int(start)           
        try: bin_dict[chr][int(pos/bin_size)] += 1
        except (IndexError, KeyError) as e: pass # index out of range at the end of chr
    
            
    return bin_dict

def parse_bam_to_bin(filename, bin_size, bin_dict, input_dir):
    num = 0
    infile = pysam.Samfile(input_dir + filename, 'rb')
    for line in infile.fetch(until_eof=True):
        num += 1
        if num %10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        if line.is_unmapped is False:
            chr = infile.getrname(line.tid)
            try: 
                bin_dict[chr][int(line.pos/BIN)] += 1
            except (IndexError, KeyError) as e: pass # index out of range at the end of chr. 
    return bin_dict           
    
def parse_sam_to_bin(filename, bin_size, bin_dict, input_dir):
    infile = open(input_dir + filename, 'r')
    # skip the header of the SAM file. 
    for line in infile:
        if not line.startswith("@"):
            break
    # start reading the real data
    num = 0
    for line in infile:
        num += 1
        if num %10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        words = line.strip().split()
        flag = int(words[1])
    
        if not flag & 0x0004: #if not unmapped
            chr, pos =  words[2], int(words[3])-1
            try: 
                bin_dict[chr][int(pos/bin_size)] += 1
            except (IndexError, KeyError) as e: pass # index out of range at end of chr.
                
    return bin_dict 
    
def parse_sampe_to_bin(filename, bin_size, bin_dict, input_dir):
    infile = open(input_dir+filename, 'r')
    num = 0
    reads_dict = {}
    flen_dict = {}
    flen_list = array.array('i',[])
    # skip the header of the SAM file.
    for line in infile:
        if not line.startswith("@"):
            pre_name = line.strip().split()[0]
            break
    line_saved = False
    # start reading the real data
    for line in infile:
        num += 1
        if num % 10000000 == 0:
            print("{0:,} lines processed in {1}".format(num, filename))
        words = line.strip().split()
        name = words[0]
        # if the sequence name has already been processed
        if name == pre_name and line_saved == True:
                continue
        # else
        #initialize the condition
        pre_name = name
        line_saved = False
        # process the new sequence. 
        flag = int(words[1])
        if not flag & 0x0004: #if not unmapped
            chr, pos,flen = words[2], int(words[3])-1, int(words[8])
            if flen != 0: # if the other end mapped to the same chromosome
                try:
                    reads_dict[chr].append(pos+flen/2)
                    flen_dict[chr].append(flen)
                except KeyError:
                    reads_dict[chr] = array.array('i',[pos+flen/2])
                    flen_dict[chr] = array.array('i', [flen])
                flen_list.append(abs(flen))
                line_saved = True

    print(num)
    # will calculate the median fragment size and remove the reads that have larger fragment size than it. 
    flen_median = numpy.median(flen_list)
    # print flen_median

    for chr in reads_dict:
        flen_chr = numpy.array(flen_dict[chr])
        reads_chr = numpy.array(reads_dict[chr])
        reads_chr = reads_chr[numpy.where(numpy.abs(flen_chr) <= 2*flen_median)[0]]
        for pos in reads_chr:
            try:
                bin_dict[chr][int(pos/bin_size)] += 1
            except (IndexError, KeyError) as e: pass # index out of range
    
    return bin_dict

def parse_bampe_to_bin(filename, bin_size, bin_dict, input_dir):
    '''parse paired-end bam file'''

    infile =pysam.Samfile(input_dir+filename, 'rb')
    num = 0
    reads_dict = {}
    flen_dict = {}
    flen_list = array.array('i',[])
    # start proccessing the data.
    line_saved = False
    pre_name = ''
    for line in infile.fetch(until_eof = True):
        num += 1
        if num % 10000000 == 0 :
            print ("{0:,} lines processed in {1}".format(num, filename))
        name = line.query_name
        if name==pre_name and line_saved == True:
            continue
        # else 
        #initialize the condition
        pre_name = name
        line_saved = False
        if line.is_unmapped is False:
            chr = infile.getrname(line.tid)
            # pos = line.pos
            # flen = line.tlen
            if line.tlen !=0:
                try: 
                    reads_dict[chr].append(line.pos+line.tlen/2)
                    flen_dict[chr].append(line.tlen)
                except KeyError:
                    reads_dict[chr] = array.array('i',[line.pos+line.tlen/2])
                    flen_dict[chr] = array.array('i',[line.tlen])
                flen_list.append(abs(line.tlen))
                line_saved = True
    print(num)
    # calculate the median fragment size and remove reads that are two times larger.
    flen_median = numpy.median(flen_list)
    for chr in reads_dict:
        flen_chr = numpy.array(flen_dict[chr])
        reads_chr = numpy.array(reads_dict[chr])
        reads_chr = reads_chr[numpy.where(numpy.abs(flen_chr) <= 2*flen_median)[0]]
        for pos in reads_chr:
            try:
                bin_dict[chr][int(pos/bin_size)] += 1
            except (IndexError, KeyError) as e: pass # index out of range

    return bin_dict
    
def parse_to_bin(filename, bin_size, parameter):
    bin_dict = {}
    for chr in parameter.chr_info:
        row_num = int(parameter.chr_info[chr]/bin_size)
        bin_dict[chr] = numpy.zeros(row_num, dtype=numpy.float64)
        
    parse_to_bin_dict = {'bed':parse_bed_to_bin,'bam':parse_bam_to_bin,'sam':parse_sam_to_bin,'sampe':parse_sampe_to_bin, 'bampe':parse_bampe_to_bin}

    bin_dict = parse_to_bin_dict[parameter.file_format](filename, bin_size, bin_dict, parameter.input_directory)

    for chr in parameter.chr_info:
        try: 
            bin_array = numpy.append(bin_array, bin_dict[chr])
        except UnboundLocalError: # if bin_array does not exist.
            bin_array = bin_dict[chr]  
            
    return bin_array
    



from optparse import OptionParser

from .classDef import Parameters


def opt_parser(argv):
    usage = ["Basic usage: %prog -c chip_files -i input_files -f file_type",
             "You could also use a parameter file by: %prog -p param_file",
             "For pre-processing only (parameter estimation), run: PePr-preprocess",
             "For post-processing, run: PePr-postprocess"]
    usage = '\n'.join(usage)
    parser = OptionParser(usage=usage)
    parser.add_option(
            "-p", "--parameter-file", action="store", type="string",
            dest="parameter", default="",
            help="provide a file that contain the parameters",
            metavar="PARAMETER-FILE")
    parser.add_option(
            "-c", "--chip1", action="store", type="string", dest="chip1",
            default="", help="chip1 file names separated by comma",
            metavar="CHIP1")
    parser.add_option(
            "-i", "--input1", action="store",
            type="string", dest="input1", default="",
            help="input1 file names separated by comma",
            metavar="INPUT1")
    parser.add_option(
            "--chip2", action="store", type="string", dest="chip2",
            default="", help="chip2 file names separated by comma",
            metavar="CHIP2")
    parser.add_option(
            "--input2", action="store", type="string", dest="input2",
            default="", help="input2 file names separated by comma",
            metavar="INPUT2")
    parser.add_option(
            "-f", "--file-format", action="store", type="string",
            dest="file_format",default="NA",
            help="bed, sam, bam, sampe, bampe...",
            metavar="FORMAT")
    parser.add_option(
            "-s", "--shiftsize", action="store",
            type="int", dest="shift_size", default=-1,
            help="Half the fragment size.", metavar="SHIFTSIZE")
    parser.add_option(
            "-w", "--windowsize", action="store",
            type="int", dest="window_size", default=-1,
            help="Window sizes",
            metavar="WINDOWSIZE")
    parser.add_option(
            "--diff", action="store_true",
            dest = "difftest", default=False,
            help="Perform differential binding instead of peak-calling")
    parser.add_option(
            "-n", "--name", action = "store",
            type="string", dest="name", default = "NA",
            help = "the experimental name. NA if none provided",
            metavar="NAME")
    parser.add_option(
            "--threshold", action ="store",
            type='float', dest="threshold", default=1e-5,
            help="p-value threshold. Default 1e-5.")
    parser.add_option(
            "--peaktype", action="store",
            type="string", dest="peaktype", default="broad",
            help="sharp or broad. Default broad.")
    parser.add_option(
            "--num-processors", action="store",
            type="int", dest="num_procs", default =1,
            help="number of cores for use.")
    parser.add_option(
            "--input-directory", action="store",
            type="string", dest="input_directory", default='',
            help="where the data files are. Absolute path recommended.")
    parser.add_option(
            "--output-directory", action="store",
            type="string", dest="output_directory", default='',
            help="where you want the output files to be")
    parser.add_option(
            "--normalization", action="store",
            type="string", dest="normalization", default="", 
            help='''Normalization method. inter-group, intra-group, scale or no. Must manually specify for differential binding analysis.''')
    parser.add_option(
            "--keep-max-dup", action="store",
            type="int", dest="keep_max_dup", default=-1,
            help='''maximum number of reads to keep at each position.
            if not specified, will not remove any duplicate.''')
    '''parser.add_option(
            "--no_log", action="store_true",
            dest = "unsave_log", default=False,
            help = "Disable saving the log files")
    '''
    parser.add_option(
            "--version", action="store_true",
            dest="version",default=False,
            help="Show version information and exit")

    (opt, args)=parser.parse_args(argv)
    if len(argv)==1:
        parser.print_help()
        exit(1)
    if opt.version == True:
        from  .__init__ import __version__
        print ("PePr "+ __version__)
        exit(1)

    return opt

def process_opt(opt):
    ''' validate the parameters that the user specified'''
    # initial process the filenames.
    opt.chip1 = [] if opt.chip1 is '' else opt.chip1.strip().split(',') 
    opt.chip2 = [] if opt.chip2 is '' else opt.chip2.strip().split(',')
    opt.input1 = [] if opt.input1 is '' else opt.input1.strip().split(',')
    opt.input2 = [] if opt.input2 is '' else opt.input2.strip().split(',')
    opt.normalization = opt.normalization.lower()
    parameter = Parameters(opt)

    #add shift size validations
    return parameter
